# Standard library imports
import base64
import os

# Third party imports
from werkzeug.datastructures import FileStorage

# Local application imports
from src.opengeodeweb_back import test_utils


def test_allowed_files(client):
    route = f"/allowed_files"
    get_full_data = lambda: {"supported_feature": "None"}
    json = get_full_data()
    response = client.post(route, json=json)
    assert response.status_code == 200
    extensions = response.json["extensions"]
    assert type(extensions) is list
    for extension in extensions:
        assert type(extension) is str

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_allowed_objects(client):
    route = f"/allowed_objects"

    def get_full_data():
        return {
            "filename": "corbi.og_brep",
            "supported_feature": None,
        }

    # Normal test with filename 'corbi.og_brep'
    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    allowed_objects = response.json["allowed_objects"]
    assert type(allowed_objects) is dict
    for allowed_object in allowed_objects:
        assert type(allowed_object) is str

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_upload_file(client):
    response = client.put(
        f"/upload_file",
        data={"file": FileStorage(open("./tests/corbi.og_brep", "rb"))},
    )

    assert response.status_code == 201


def test_missing_files(client):
    route = f"/missing_files"

    def get_full_data():
        return {
            "input_geode_object": "BRep",
            "filename": "corbi.og_brep",
        }

    json = get_full_data()
    response = client.post(route, json=json)
    assert response.status_code == 200
    has_missing_files = response.json["has_missing_files"]
    mandatory_files = response.json["mandatory_files"]
    additional_files = response.json["additional_files"]
    assert type(has_missing_files) is bool
    assert type(mandatory_files) is list
    assert type(additional_files) is list

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_geographic_coordinate_systems(client):
    route = f"/geographic_coordinate_systems"
    get_full_data = lambda: {"input_geode_object": "BRep"}
    # Normal test with geode_object 'BRep'
    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    crs_list = response.json["crs_list"]
    assert type(crs_list) is list
    for crs in crs_list:
        assert type(crs) is dict

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_inspect_file(client):
    route = f"/inspect_file"

    def get_full_data():
        return {
            "input_geode_object": "BRep",
            "filename": "corbi.og_brep",
        }

    json = get_full_data()

    # Normal test with geode_object 'BRep'
    response = client.post(route, json=json)
    assert response.status_code == 200
    inspection_result = response.json["inspection_result"]
    assert type(inspection_result) is dict

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_geode_objects_and_output_extensions(client):
    route = "/geode_objects_and_output_extensions"

    def get_full_data():
        return {
            "input_geode_object": "BRep",
            "filename": "corbi.og_brep",
        }

    response = client.post(route, json=get_full_data())

    assert response.status_code == 200
    geode_objects_and_output_extensions = response.json[
        "geode_objects_and_output_extensions"
    ]
    assert type(geode_objects_and_output_extensions) is dict
    for geode_object, values in geode_objects_and_output_extensions.items():
        assert type(values) is dict
        for output_extension, value in values.items():
            assert type(value) is dict
            assert type(value["is_saveable"]) is bool

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_save_viewable_file(client):
    route = f"/save_viewable_file"

    def get_full_data():
        return {
            "input_geode_object": "BRep",
            "filename": "corbi.og_brep",
        }

    # Normal test with filename 'corbi.og_brep'
    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    name = response.json["name"]
    assert type(name) is str
    native_file_name = response.json["native_file_name"]
    assert type(native_file_name) is str
    viewable_file_name = response.json["viewable_file_name"]
    assert type(viewable_file_name) is str
    id = response.json["id"]
    assert type(id) is str
    object_type = response.json["object_type"]
    assert type(object_type) is str
    assert object_type in ["model", "mesh"]

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_vertex_attribute_names(client):
    response = client.put(
        f"/upload_file",
        data={"file": FileStorage(open("./tests/vertex_attribute.vtp", "rb"))},
    )
    assert response.status_code == 201

    route = f"/vertex_attribute_names"

    def get_full_data():
        return {
            "input_geode_object": "PolygonalSurface3D",
            "filename": "vertex_attribute.vtp",
        }

    # Normal test with filename 'vertex_attribute.vtp'
    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    vertex_attribute_names = response.json["vertex_attribute_names"]
    assert type(vertex_attribute_names) is list
    for vertex_attribute_name in vertex_attribute_names:
        assert type(vertex_attribute_name) is str

    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_polygon_attribute_names(client):
    response = client.put(
        f"/upload_file",
        data={"file": FileStorage(open("./tests/polygon_attribute.vtp", "rb"))},
    )
    assert response.status_code == 201

    route = f"/polygon_attribute_names"

    def get_full_data():
        return {
            "input_geode_object": "PolygonalSurface3D",
            "filename": "polygon_attribute.vtp",
        }

    # Normal test with filename 'vertex_attribute.vtp'
    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    polygon_attribute_names = response.json["polygon_attribute_names"]
    assert type(polygon_attribute_names) is list
    for polygon_attribute_name in polygon_attribute_names:
        assert type(polygon_attribute_name) is str

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_create_point(client):
    route = f"/create_point"
    get_full_data = lambda: {"x": 1, "y": 2, "z": 3}

    # Normal test with all keys
    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    viewable_file_name = response.json["viewable_file_name"]
    assert type(viewable_file_name) is str
    id = response.json["id"]
    assert type(id) is str

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)



# Standard library imports
import os
import threading
import time
import zipfile

# Third party imports
import flask
from jsonschema import validate
from jsonschema.exceptions import ValidationError
import pkg_resources

# Local application imports


def increment_request_counter(current_app):
    if "REQUEST_COUNTER" in current_app.config:
        REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER"))
        REQUEST_COUNTER += 1
        current_app.config.update(REQUEST_COUNTER=REQUEST_COUNTER)


def decrement_request_counter(current_app):
    if "REQUEST_COUNTER" in current_app.config:
        REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER"))
        REQUEST_COUNTER -= 1
        current_app.config.update(REQUEST_COUNTER=REQUEST_COUNTER)


def update_last_request_time(current_app):
    if "LAST_REQUEST_TIME" in current_app.config:
        LAST_REQUEST_TIME = time.time()
        current_app.config.update(LAST_REQUEST_TIME=LAST_REQUEST_TIME)


def before_request(current_app):
    increment_request_counter(current_app)


def teardown_request(current_app):
    decrement_request_counter(current_app)
    update_last_request_time(current_app)


def kill_task(current_app):
    DESKTOP_APP = bool(current_app.config.get("DESKTOP_APP"))
    REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER"))
    LAST_PING_TIME = float(current_app.config.get("LAST_PING_TIME"))
    LAST_REQUEST_TIME = float(current_app.config.get("LAST_REQUEST_TIME"))
    MINUTES_BEFORE_TIMEOUT = float(current_app.config.get("MINUTES_BEFORE_TIMEOUT"))
    current_time = time.time()
    minutes_since_last_request = (current_time - LAST_REQUEST_TIME) / 60
    minutes_since_last_ping = (current_time - LAST_PING_TIME) / 60

    if (
        (
            (minutes_since_last_request > MINUTES_BEFORE_TIMEOUT)
            and (DESKTOP_APP == False)
        )
        or (minutes_since_last_ping > MINUTES_BEFORE_TIMEOUT)
    ) and (REQUEST_COUNTER == 0):
        print("Server timed out due to inactivity, shutting down...", flush=True)
        os._exit(0)


def versions(list_packages: list):
    list_with_versions = []
    for package in list_packages:
        list_with_versions.append(
            {
                "package": package,
                "version": pkg_resources.get_distribution(package).version,
            }
        )
    return list_with_versions


def validate_request(request, schema):
    json_data = request.get_json(force=True, silent=True)

    if json_data is None:
        json_data = {}

    try:
        validate(instance=json_data, schema=schema)
    except ValidationError as e:
        flask.abort(400, f"Validation error: {e.message}")


def set_interval(func, sec, args=None):
    def func_wrapper():
        set_interval(func, sec, args)
        func(args)

    t = threading.Timer(sec, func_wrapper)
    t.daemon = True
    t.start()
    return t


def extension_from_filename(filename):
    return os.path.splitext(filename)[1][1:]


def send_file(upload_folder, saved_files, new_file_name):
    if len(saved_files) == 1:
        mimetype = "application/octet-binary"
    else:
        mimetype = "application/zip"
        new_file_name = os.path.splitext(new_file_name)[0] + ".zip"
        with zipfile.ZipFile(os.path.join(upload_folder, new_file_name), "w") as zipObj:
            for saved_file_path in saved_files:
                zipObj.write(
                    saved_file_path,
                    os.path.basename(saved_file_path),
                )

    response = flask.send_from_directory(
        directory=upload_folder,
        path=new_file_name,
        as_attachment=True,
        mimetype=mimetype,
    )
    response.headers["new-file-name"] = new_file_name
    response.headers["Access-Control-Expose-Headers"] = "new-file-name"

    return response


def handle_exception(e):
    response = e.get_response()
    response.data = flask.json.dumps(
        {
            "code": e.code,
            "name": e.name,
            "description": e.description,
        }
    )
    response.content_type = "application/json"
    return response



# Standard library imports

# Third party imports

# Local application imports


def test_route_wrong_params(client, route, get_full_data):
    for key, value in get_full_data().items():
        json = get_full_data()
        json.pop(key)
        response = client.post(route, json=json)
        assert response.status_code == 400
        error_description = response.json["description"]
        assert error_description == f"Validation error: '{key}' is a required property"

    json = get_full_data()
    json["dumb_key"] = "dumb_value"
    response = client.post(route, json=json)
    assert response.status_code == 400
    error_description = response.json["description"]
    assert (
        error_description
        == "Validation error: Additional properties are not allowed ('dumb_key' was unexpected)"
    )



# Standard library imports
import os

# Third party imports
import flask
import opengeode_geosciences as og_gs
import opengeode as og

# Local application imports
from .geode_objects import geode_objects_dict
from . import utils_functions


def geode_object_value(geode_object: str):
    return geode_objects_dict()[geode_object]


def geode_object_class(geode_object: str):
    return geode_object_value(geode_object)["class"]


def input_factory(geode_object: str):
    return geode_object_value(geode_object)["input_factory"]


def output_factory(geode_object: str):
    return geode_object_value(geode_object)["output_factory"]


def missing_files(geode_object: str, file_absolute_path: str):
    return geode_object_value(geode_object)["missing_files"](file_absolute_path)


def is_loadable(geode_object: str, file_absolute_path: str):
    return geode_object_value(geode_object)["is_loadable"](file_absolute_path)


def load(geode_object: str, file_absolute_path: str):
    return geode_object_value(geode_object)["load"](file_absolute_path)


def is_saveable(geode_object: str, data, filename: str):
    return geode_object_value(geode_object)["is_saveable"](data, filename)


def save(geode_object: str, data, folder_absolute_path: str, filename: str):
    return geode_object_value(geode_object)["save"](
        data, os.path.join(folder_absolute_path, filename)
    )


def create_builder(geode_object: str, data):
    return geode_object_value(geode_object)["builder"](data)


def assign_crs(geode_object: str, data, crs_name: str, info):
    builder = create_builder(geode_object, data)
    geode_object_value(geode_object)["crs"]["assign"](data, builder, crs_name, info)


def convert_crs(geode_object: str, data, crs_name: str, info):
    builder = create_builder(geode_object, data)
    geode_object_value(geode_object)["crs"]["convert"](data, builder, crs_name, info)


def create_crs(
    geode_object: str,
    data,
    name: str,
    input_coordiante_system,
    output_coordiante_system,
):
    builder = create_builder(geode_object, data)
    geode_object_value(geode_object)["crs"]["create"](
        data, builder, name, input_coordiante_system, output_coordiante_system
    )


def get_object_type(geode_object: str):
    return geode_object_value(geode_object)["object_type"]


def is_3D(geode_object: str):
    return geode_object_value(geode_object)["is_3D"]


def is_viewable(geode_object: str):
    return geode_object_value(geode_object)["is_viewable"]


def inspect(geode_object: str, data):
    return geode_object_value(geode_object)["inspector"](data)


def save_viewable(geode_object: str, data, folder_absolute_path: str, id: str):
    return geode_object_value(geode_object)["save_viewable"](
        data, os.path.join(folder_absolute_path, id)
    )


def geode_object_input_extensions(geode_object: str):
    geode_object_input_list_creators = input_factory(geode_object).list_creators()
    geode_object_input_list_creators.sort()
    return geode_object_input_list_creators


def geode_object_output_extensions(geode_object: str):
    geode_object_output_list_creators = output_factory(geode_object).list_creators()
    geode_object_output_list_creators.sort()
    return geode_object_output_list_creators


def filter_geode_objects(key: str = None):
    geode_objects_filtered_list = []
    for geode_object, value in geode_objects_dict().items():
        if key != None and key != "":
            if key in value:
                if type(value[key]) == bool:
                    geode_objects_filtered_list.append(geode_object)
                else:
                    geode_objects_filtered_list.append(geode_object)
        else:
            geode_objects_filtered_list.append(geode_object)
    geode_objects_filtered_list.sort()
    return geode_objects_filtered_list


def list_input_extensions(key: str = None):
    extensions_list = []
    geode_objects_filtered_list = filter_geode_objects(key)
    for geode_object in geode_objects_filtered_list:
        extensions_list += geode_object_input_extensions(geode_object)

    extensions_list = list(set(extensions_list))
    extensions_list.sort()
    return extensions_list


def has_creator(geode_object: str, extension: str):
    return input_factory(geode_object).has_creator(extension)


def list_geode_objects(
    file_absolute_path: str,
    key: str = None,
):
    return_dict = {}
    file_extension = utils_functions.extension_from_filename(
        os.path.basename(file_absolute_path)
    )
    geode_objects_filtered_list = filter_geode_objects(key)

    for geode_object in geode_objects_filtered_list:
        if has_creator(geode_object, file_extension):
            file_is_loadable = is_loadable(geode_object, file_absolute_path)
            return_dict[geode_object] = {"is_loadable": file_is_loadable}
    return return_dict


def geode_objects_output_extensions(geode_object: str, data):
    geode_objects_output_extensions_dict = {}
    output_extensions = geode_object_output_extensions(geode_object)
    extensions_dict = {}
    for output_extension in output_extensions:
        bool_is_saveable = is_saveable(geode_object, data, f"test.{output_extension}")
        extensions_dict[output_extension] = {"is_saveable": bool_is_saveable}
    geode_objects_output_extensions_dict[geode_object] = extensions_dict

    if "parent" in geode_object_value(geode_object).keys():
        parent_geode_object = geode_object_value(geode_object)["parent"]
        geode_objects_output_extensions_dict.update(
            geode_objects_output_extensions(parent_geode_object, data)
        )
    return geode_objects_output_extensions_dict


def get_inspector_children(obj):
    new_object = {}

    if "inspection_type" in dir(obj):
        new_object["title"] = obj.inspection_type()
        new_object["nb_issues"] = 0
        new_object["children"] = []
        for child in dir(obj):
            if not child.startswith("__") and not child in [
                "inspection_type",
                "string",
            ]:
                child_instance = obj.__getattribute__(child)
                child_object = get_inspector_children(child_instance)
                new_object["children"].append(child_object)
                new_object["nb_issues"] += child_object["nb_issues"]
    else:
        new_object["title"] = obj.description()
        nb_issues = obj.nb_issues()
        new_object["nb_issues"] = nb_issues
        if nb_issues > 0:
            issues = obj.string().split("\n")
            new_object["issues"] = issues
    return new_object


def geographic_coordinate_systems(geode_object: str):
    if is_3D(geode_object):
        return og_gs.GeographicCoordinateSystem3D.geographic_coordinate_systems()
    else:
        return og_gs.GeographicCoordinateSystem2D.geographic_coordinate_systems()


def geographic_coordinate_systems_info(geode_object: str, crs):
    if is_3D(geode_object):
        return og_gs.GeographicCoordinateSystemInfo3D(
            crs["authority"], crs["code"], crs["name"]
        )
    else:
        return og_gs.GeographicCoordinateSystemInfo2D(
            crs["authority"], crs["code"], crs["name"]
        )


def coordinate_system(geode_object: str, coordinate_system):
    return og.CoordinateSystem2D(
        [
            og.Vector2D(
                og.Point2D(
                    [coordinate_system["origin_x"], coordinate_system["origin_y"]]
                ),
                og.Point2D(
                    [coordinate_system["point_1_x"], coordinate_system["point_1_y"]]
                ),
            ),
            og.Vector2D(
                og.Point2D(
                    [coordinate_system["origin_x"], coordinate_system["origin_y"]]
                ),
                og.Point2D(
                    [coordinate_system["point_2_x"], coordinate_system["point_2_y"]]
                ),
            ),
        ],
        og.Point2D([coordinate_system["origin_x"], coordinate_system["origin_y"]]),
    )


def assign_geographic_coordinate_system_info(geode_object: str, data, input_crs):
    info = geographic_coordinate_systems_info(geode_object, input_crs)
    assign_crs(geode_object, data, input_crs["name"], info)


def convert_geographic_coordinate_system_info(geode_object: str, data, output_crs):
    info = geographic_coordinate_systems_info(geode_object, output_crs)
    convert_crs(geode_object, data, output_crs["name"], info)


def create_coordinate_system(
    geode_object: str, data, name, input_coordinate_points, output_coordinate_points
):
    input_coordiante_system = coordinate_system(geode_object, input_coordinate_points)
    output_coordiante_system = coordinate_system(geode_object, output_coordinate_points)
    create_crs(
        geode_object, data, name, input_coordiante_system, output_coordiante_system
    )



# Standard library imports
import json
import os
import time

# Third party imports
import flask
import opengeode
import uuid
import werkzeug

# Local application imports
from .. import geode_functions, utils_functions

routes = flask.Blueprint("routes", __name__)


@routes.before_request
def before_request():
    if "ping" not in flask.request.path:
        utils_functions.increment_request_counter(flask.current_app)


@routes.teardown_request
def teardown_request(exception):
    if "ping" not in flask.request.path:
        utils_functions.decrement_request_counter(flask.current_app)
        utils_functions.update_last_request_time(flask.current_app)


schemas = os.path.join(os.path.dirname(__file__), "schemas")

with open(
    os.path.join(schemas, "allowed_files.json"),
    "r",
) as file:
    allowed_files_json = json.load(file)


@routes.route(
    allowed_files_json["route"],
    methods=allowed_files_json["methods"],
)
def allowed_files():
    utils_functions.validate_request(flask.request, allowed_files_json)
    extensions = geode_functions.list_input_extensions(
        flask.request.json["supported_feature"]
    )
    return flask.make_response({"extensions": extensions}, 200)


with open(
    os.path.join(schemas, "upload_file.json"),
    "r",
) as file:
    upload_file_json = json.load(file)


@routes.route(
    upload_file_json["route"],
    methods=upload_file_json["methods"],
)
def upload_file():
    if flask.request.method == "OPTIONS":
        return flask.make_response({}, 200)

    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    if not os.path.exists(UPLOAD_FOLDER):
        os.mkdir(UPLOAD_FOLDER)
    file = flask.request.files["file"]
    filename = werkzeug.utils.secure_filename(os.path.basename(file.filename))
    file.save(os.path.join(UPLOAD_FOLDER, filename))
    return flask.make_response({"message": "File uploaded"}, 201)


with open(
    os.path.join(schemas, "allowed_objects.json"),
    "r",
) as file:
    allowed_objects_json = json.load(file)


@routes.route(
    allowed_objects_json["route"],
    methods=allowed_objects_json["methods"],
)
def allowed_objects():
    if flask.request.method == "OPTIONS":
        return flask.make_response({}, 200)

    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    utils_functions.validate_request(flask.request, allowed_objects_json)
    file_absolute_path = os.path.join(UPLOAD_FOLDER, flask.request.json["filename"])
    allowed_objects = geode_functions.list_geode_objects(
        file_absolute_path, flask.request.json["supported_feature"]
    )
    return flask.make_response({"allowed_objects": allowed_objects}, 200)


with open(
    os.path.join(schemas, "missing_files.json"),
    "r",
) as file:
    missing_files_json = json.load(file)


@routes.route(
    missing_files_json["route"],
    methods=missing_files_json["methods"],
)
def missing_files():
    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    utils_functions.validate_request(flask.request, missing_files_json)

    missing_files = geode_functions.missing_files(
        flask.request.json["input_geode_object"],
        os.path.join(UPLOAD_FOLDER, flask.request.json["filename"]),
    )
    has_missing_files = missing_files.has_missing_files()

    mandatory_files = []
    for mandatory_file in missing_files.mandatory_files:
        mandatory_files.append(os.path.basename(mandatory_file))

    additional_files = []
    for additional_file in missing_files.additional_files:
        additional_files.append(os.path.basename(additional_file))

    return flask.make_response(
        {
            "has_missing_files": has_missing_files,
            "mandatory_files": mandatory_files,
            "additional_files": additional_files,
        },
        200,
    )


with open(
    os.path.join(schemas, "geographic_coordinate_systems.json"),
    "r",
) as file:
    geographic_coordinate_systems_json = json.load(file)


@routes.route(
    geographic_coordinate_systems_json["route"],
    methods=geographic_coordinate_systems_json["methods"],
)
def crs_converter_geographic_coordinate_systems():
    utils_functions.validate_request(flask.request, geographic_coordinate_systems_json)
    infos = geode_functions.geographic_coordinate_systems(
        flask.request.json["input_geode_object"]
    )
    crs_list = []
    for info in infos:
        crs = {}
        crs["name"] = info.name
        crs["code"] = info.code
        crs["authority"] = info.authority
        crs_list.append(crs)

    return flask.make_response({"crs_list": crs_list}, 200)


with open(
    os.path.join(schemas, "inspect_file.json"),
    "r",
) as file:
    inspect_file_json = json.load(file)


@routes.route(
    inspect_file_json["route"],
    methods=inspect_file_json["methods"],
)
def inspect_file():
    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    utils_functions.validate_request(flask.request, inspect_file_json)

    secure_filename = werkzeug.utils.secure_filename(flask.request.json["filename"])
    file_path = os.path.abspath(os.path.join(UPLOAD_FOLDER, secure_filename))
    data = geode_functions.load(flask.request.json["input_geode_object"], file_path)
    class_inspector = geode_functions.inspect(
        flask.request.json["input_geode_object"], data
    )
    inspection_result = geode_functions.get_inspector_children(class_inspector)
    return flask.make_response({"inspection_result": inspection_result}, 200)


with open(
    os.path.join(schemas, "geode_objects_and_output_extensions.json"),
    "r",
) as file:
    geode_objects_and_output_extensions_json = json.load(file)


@routes.route(
    geode_objects_and_output_extensions_json["route"],
    methods=geode_objects_and_output_extensions_json["methods"],
)
def geode_objects_and_output_extensions():
    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    utils_functions.validate_request(
        flask.request, geode_objects_and_output_extensions_json
    )
    data = geode_functions.load(
        flask.request.json["input_geode_object"],
        os.path.join(UPLOAD_FOLDER, flask.request.json["filename"]),
    )
    geode_objects_and_output_extensions = (
        geode_functions.geode_objects_output_extensions(
            flask.request.json["input_geode_object"], data
        )
    )
    return flask.make_response(
        {"geode_objects_and_output_extensions": geode_objects_and_output_extensions},
        200,
    )


with open(
    os.path.join(schemas, "save_viewable_file.json"),
    "r",
) as file:
    save_viewable_file_json = json.load(file)


@routes.route(
    save_viewable_file_json["route"],
    methods=save_viewable_file_json["methods"],
)
def save_viewable_file():
    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    DATA_FOLDER_PATH = flask.current_app.config["DATA_FOLDER_PATH"]
    utils_functions.validate_request(flask.request, save_viewable_file_json)

    secure_filename = werkzeug.utils.secure_filename(flask.request.json["filename"])
    file_path = os.path.abspath(os.path.join(UPLOAD_FOLDER, secure_filename))
    data = geode_functions.load(flask.request.json["input_geode_object"], file_path)
    generated_id = str(uuid.uuid4()).replace("-", "")

    if geode_functions.is_viewable(flask.request.json["input_geode_object"]):
        name = data.name()
    else:
        name = flask.request.json["filename"]

    native_extension = data.native_extension()

    absolute_native_file_path = os.path.join(
        UPLOAD_FOLDER, generated_id + "." + native_extension
    )

    saved_viewable_file_path = geode_functions.save_viewable(
        flask.request.json["input_geode_object"], data, DATA_FOLDER_PATH, generated_id
    )
    geode_functions.save(
        flask.request.json["input_geode_object"],
        data,
        DATA_FOLDER_PATH,
        generated_id + "." + native_extension,
    )
    object_type = geode_functions.get_object_type(
        flask.request.json["input_geode_object"]
    )

    native_file_name = os.path.basename(absolute_native_file_path)
    viewable_file_name = os.path.basename(saved_viewable_file_path)
    return flask.make_response(
        {
            "name": name,
            "native_file_name": native_file_name,
            "viewable_file_name": viewable_file_name,
            "id": generated_id,
            "object_type": object_type,
        },
        200,
    )


with open(os.path.join(schemas, "create_point.json"), "r") as file:
    create_point_json = json.load(file)


@routes.route(create_point_json["route"], methods=create_point_json["methods"])
def create_point():
    utils_functions.validate_request(flask.request, create_point_json)
    DATA_FOLDER_PATH = flask.current_app.config["DATA_FOLDER_PATH"]
    x = flask.request.json["x"]
    y = flask.request.json["y"]
    z = flask.request.json["z"]
    class_ = geode_functions.geode_object_class("PointSet3D")
    PointSet3D = class_.create()
    builder = geode_functions.create_builder("PointSet3D", PointSet3D)
    builder.create_point(opengeode.Point3D([x, y, z]))

    generated_id = str(uuid.uuid4()).replace("-", "")
    saved_viewable_file_path = geode_functions.save_viewable(
        "PointSet3D", PointSet3D, DATA_FOLDER_PATH, generated_id
    )

    return flask.make_response(
        {
            "viewable_file_name": os.path.basename(saved_viewable_file_path),
            "id": generated_id,
        },
        200,
    )


with open(
    os.path.join(schemas, "vertex_attribute_names.json"),
    "r",
) as file:
    vertex_attribute_names_json = json.load(file)


@routes.route(
    vertex_attribute_names_json["route"],
    methods=vertex_attribute_names_json["methods"],
)
def vertex_attribute_names():

    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    utils_functions.validate_request(flask.request, vertex_attribute_names_json)
    file_absolute_path = os.path.join(
        UPLOAD_FOLDER, werkzeug.utils.secure_filename(flask.request.json["filename"])
    )
    data = geode_functions.load(
        flask.request.json["input_geode_object"], file_absolute_path
    )
    vertex_attribute_names = data.vertex_attribute_manager().attribute_names()

    return flask.make_response(
        {
            "vertex_attribute_names": vertex_attribute_names,
        },
        200,
    )


with open(
    os.path.join(schemas, "polygon_attribute_names.json"),
    "r",
) as file:
    polygon_attribute_names_json = json.load(file)


@routes.route(
    polygon_attribute_names_json["route"],
    methods=polygon_attribute_names_json["methods"],
)
def polygon_attribute_names():

    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    utils_functions.validate_request(flask.request, vertex_attribute_names_json)
    file_absolute_path = os.path.join(
        UPLOAD_FOLDER, werkzeug.utils.secure_filename(flask.request.json["filename"])
    )
    data = geode_functions.load(
        flask.request.json["input_geode_object"], file_absolute_path
    )
    polygon_attribute_names = data.polygon_attribute_manager().attribute_names()

    return flask.make_response(
        {
            "polygon_attribute_names": polygon_attribute_names,
        },
        200,
    )


with open(
    os.path.join(schemas, "ping.json"),
    "r",
) as file:
    ping_json = json.load(file)


@routes.route(
    ping_json["route"],
    methods=ping_json["methods"],
)
def ping():
    utils_functions.validate_request(flask.request, ping_json)
    flask.current_app.config.update(LAST_PING_TIME=time.time())
    return flask.make_response({"message": "Flask server is running"}, 200)



# Standard library imports

# Third party imports
import flask

# Local application imports
from src.opengeodeweb_back import utils_functions


def test_increment_request_counter(app_context):
    assert flask.current_app.config.get("REQUEST_COUNTER") == 0
    utils_functions.increment_request_counter(flask.current_app)
    assert flask.current_app.config.get("REQUEST_COUNTER") == 1


def test_decrement_request_counter(app_context):
    assert flask.current_app.config.get("REQUEST_COUNTER") == 1
    utils_functions.decrement_request_counter(flask.current_app)
    assert flask.current_app.config.get("REQUEST_COUNTER") == 0


def test_update_last_request_time(app_context):
    LAST_REQUEST_TIME = flask.current_app.config.get("LAST_REQUEST_TIME")
    utils_functions.update_last_request_time(flask.current_app)
    assert flask.current_app.config.get("LAST_REQUEST_TIME") >= LAST_REQUEST_TIME


def test_before_request(app_context):
    assert flask.current_app.config.get("REQUEST_COUNTER") == 0
    utils_functions.before_request(flask.current_app)
    assert flask.current_app.config.get("REQUEST_COUNTER") == 1


def test_teardown_request(app_context):
    LAST_REQUEST_TIME = flask.current_app.config.get("LAST_REQUEST_TIME")
    assert flask.current_app.config.get("REQUEST_COUNTER") == 1
    utils_functions.teardown_request(flask.current_app)
    assert flask.current_app.config.get("REQUEST_COUNTER") == 0
    assert flask.current_app.config.get("LAST_REQUEST_TIME") >= LAST_REQUEST_TIME


def test_versions():
    list_packages = [
        "OpenGeode-core",
        "OpenGeode-IO",
        "OpenGeode-Geosciences",
        "OpenGeode-GeosciencesIO",
    ]
    versions = utils_functions.versions(list_packages)
    assert type(versions) is list
    for version in versions:
        assert type(version) is dict


def test_extension_from_filename():
    extension = utils_functions.extension_from_filename("test.toto")
    assert type(extension) is str
    assert extension.count(".") == 0


def test_handle_exception(client):
    route = "/error"
    response = client.post(route)
    assert response.status_code == 500
    data = response.get_json()
    assert type(data) is dict
    assert type(data["description"]) is str
    assert type(data["name"]) is str
    assert type(data["code"]) is int



# Standard library imports
import os
import time

# Third party imports
# Local application imports


class Config(object):
    FLASK_DEBUG = os.environ.get("FLASK_DEBUG", default=False)
    DEFAULT_HOST = "localhost"
    DEFAULT_PORT = "5000"
    CORS_HEADERS = "Content-Type"
    UPLOAD_FOLDER = "./uploads"
    DESKTOP_APP = False
    REQUEST_COUNTER = 0
    LAST_REQUEST_TIME = time.time()
    LAST_PING_TIME = time.time()


class ProdConfig(Config):
    SSL = None
    ORIGINS = ""
    MINUTES_BEFORE_TIMEOUT = "1"
    SECONDS_BETWEEN_SHUTDOWNS = "10"
    DATA_FOLDER_PATH = "/data/"


class DevConfig(Config):
    SSL = None
    ORIGINS = "*"
    MINUTES_BEFORE_TIMEOUT = "1"
    SECONDS_BETWEEN_SHUTDOWNS = "10"
    DATA_FOLDER_PATH = "./data/"



# Standard library imports
import os
import uuid

# Third party imports

# Local application imports
from src.opengeodeweb_back import geode_functions, geode_objects


data_folder = os.path.join(os.path.dirname(__file__), "data")


def test_geode_object_value():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        geode_object_value = geode_functions.geode_object_value(geode_object)
        assert type(geode_object_value) is dict


def test_input_factory():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        geode_object_input = geode_functions.input_factory(geode_object)
        assert type(geode_object_input.list_creators()) is list


def test_input_extensions():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        geode_object_input_extensions = geode_functions.geode_object_input_extensions(
            geode_object
        )

        assert type(geode_object_input_extensions) is list
        for extension in geode_object_input_extensions:
            assert type(extension) is str


def test_output_factory():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        geode_object_output = geode_functions.output_factory(geode_object)
        geode_object_output_list = geode_object_output.list_creators()
        assert type(geode_object_output_list) is list
        for output in geode_object_output_list:
            assert type(output) is str


def test_missing_files():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        input_extensions = geode_functions.geode_object_input_extensions(geode_object)
        for input_extension in input_extensions:
            file_absolute_path = os.path.join(data_folder, f"test.{input_extension}")
            missing_files = geode_functions.missing_files(
                geode_object, file_absolute_path
            )
            has_missing_files = missing_files.has_missing_files()
            assert type(has_missing_files) is bool
            mandatory_files = missing_files.mandatory_files
            assert type(mandatory_files) is list
            additional_files = missing_files.additional_files
            assert type(additional_files) is list


def test_is_loadable():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        input_extensions = geode_functions.geode_object_input_extensions(geode_object)
        for input_extension in input_extensions:
            file_absolute_path = os.path.join(data_folder, f"test.{input_extension}")
            is_loadable = geode_functions.is_loadable(geode_object, file_absolute_path)
            assert type(is_loadable) is bool


def test_load():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        print(f"\n{geode_object=}")
        input_extensions = geode_functions.geode_object_input_extensions(geode_object)
        for input_extension in input_extensions:
            print(f"\t{input_extension=}")
            file_absolute_path = os.path.join(data_folder, f"test.{input_extension}")
            missing_files = geode_functions.missing_files(
                geode_object, file_absolute_path
            )
            has_missing_files = missing_files.has_missing_files()
            if has_missing_files:
                mandatory_files = missing_files.mandatory_files
                print(f"\t\t{mandatory_files=}")
                additional_files = missing_files.additional_files
                print(f"\t\t{additional_files=}")
            if geode_functions.is_loadable(geode_object, file_absolute_path):
                data = geode_functions.load(geode_object, file_absolute_path)
                if "save_viewable" in value:
                    uu_id = str(uuid.uuid4()).replace("-", "")
                    viewable_file_path = geode_functions.save_viewable(
                        geode_object,
                        data,
                        os.path.abspath(f"./output"),
                        uu_id,
                    )
                    os.remove(viewable_file_path)
                geode_objects_and_output_extensions = (
                    geode_functions.geode_objects_output_extensions(geode_object, data)
                )
                assert type(geode_objects_and_output_extensions) is dict
                for (
                    output_geode_object,
                    output_geode_object_value,
                ) in geode_objects_and_output_extensions.items():
                    print(f"\t\t{output_geode_object=}")
                    for (
                        output_extension,
                        output_extension_value,
                    ) in output_geode_object_value.items():
                        print(f"\t\t\t{output_extension=}")
                        uu_id = str(uuid.uuid4()).replace("-", "")
                        filename = f"{uu_id}.{output_extension}"
                        if geode_functions.is_saveable(
                            output_geode_object, data, filename
                        ):
                            saved_files = geode_functions.save(
                                output_geode_object,
                                data,
                                os.path.abspath(f"./output"),
                                filename,
                            )
                            assert type(saved_files) is list
                            for saved_file in saved_files:
                                os.remove(saved_file)


def test_get_object_type():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        object_type = geode_functions.get_object_type(geode_object)
        assert type(object_type) is str
        assert object_type in ["model", "mesh"]


def test_is_3D():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        is_3D = geode_functions.is_3D(geode_object)
        assert type(is_3D) is bool


def test_is_viewable():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        is_viewable = geode_functions.is_viewable(geode_object)
        assert type(is_viewable) is bool


def test_geode_object_input_extensions():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        input_extensions = geode_functions.geode_object_input_extensions(geode_object)
        assert type(input_extensions) is list
        for input_extension in input_extensions:
            assert type(input_extension) is str


def test_geode_object_output_extensions():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        print(f"\n{geode_object=}")
        input_extensions = geode_functions.geode_object_input_extensions(geode_object)
        for input_extension in input_extensions:
            print(f"\t{input_extension=}")
            file_absolute_path = os.path.join(data_folder, f"test.{input_extension}")
            missing_files = geode_functions.missing_files(
                geode_object, file_absolute_path
            )
            has_missing_files = missing_files.has_missing_files()
            if has_missing_files:
                mandatory_files = missing_files.mandatory_files
                print(f"\t\t{mandatory_files=}")
                additional_files = missing_files.additional_files
                print(f"\t\t{additional_files=}")
            file_absolute_path = os.path.join(data_folder, f"test.{input_extension}")
            if geode_functions.is_loadable(geode_object, file_absolute_path):
                data = geode_functions.load(geode_object, file_absolute_path)
                geode_objets_and_output_extensions = (
                    geode_functions.geode_objects_output_extensions(geode_object, data)
                )
                assert type(geode_objets_and_output_extensions) is dict
                for (
                    output_geode_object,
                    output_geode_object_value,
                ) in geode_objets_and_output_extensions.items():
                    for (
                        output_extension,
                        output_extension_value,
                    ) in output_geode_object_value.items():
                        assert type(output_extension) is str
                        assert type(output_extension_value["is_saveable"]) is bool


def test_get_inspector_children():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        if "inspector" in value:
            print(f"\n{geode_object=}", flush=True)
            input_extensions = geode_functions.geode_object_input_extensions(
                geode_object
            )
            for input_extension in input_extensions:
                print(f"\t{input_extension=}", flush=True)
                file_absolute_path = os.path.join(
                    data_folder, f"test.{input_extension}"
                )
                missing_files = geode_functions.missing_files(
                    geode_object, file_absolute_path
                )
                has_missing_files = missing_files.has_missing_files()
                if has_missing_files:
                    mandatory_files = missing_files.mandatory_files
                    print(f"\t\t{mandatory_files=}", flush=True)
                    additional_files = missing_files.additional_files
                    print(f"\t\t{additional_files=}", flush=True)
                file_absolute_path = os.path.join(
                    data_folder, f"test.{input_extension}"
                )
                if geode_functions.is_loadable(geode_object, file_absolute_path):
                    data = geode_functions.load(geode_object, file_absolute_path)
                    class_inspector = geode_functions.inspect(geode_object, data)
                    assert "InspectionResult" in class_inspector.__class__.__name__
                    inspection_result = geode_functions.get_inspector_children(
                        class_inspector
                    )

                    print(f"\t\t{inspection_result=}", flush=True)
                    assert type(inspection_result) is dict


def test_filter_geode_objects():
    filters_list = ["", "crs", "inspector", None]

    for filter in filters_list:
        geode_objects_filtered_list = geode_functions.filter_geode_objects(filter)
        assert type(geode_objects_filtered_list) is list
        for geode_object in geode_objects_filtered_list:
            assert type(geode_object) is str


def test_list_input_extensions():
    keys_array = ["crs", "inspector", None]
    for geode_object, value in geode_objects.geode_objects_dict().items():
        for keys in keys_array:
            input_extensions = geode_functions.list_input_extensions(keys)
            assert type(input_extensions) is list


def test_list_geode_objects():
    test_list = [
        {
            "key": "crs",
            "invalid_geode_objects": [
                "Graph",
                "RasterImage2D",
                "RasterImage3D",
                "VertexSet",
            ],
        },
        {
            "key": "inspector",
            "invalid_geode_objects": [
                "Graph",
                "RasterImage2D",
                "RasterImage3D",
                "VertexSet",
            ],
        },
        {
            "key": None,
            "invalid_geode_objects": [],
        },
    ]
    for test in test_list:
        key = test["key"]
        invalid_geode_objects = test["invalid_geode_objects"]

        input_extensions = geode_functions.list_input_extensions(key)
        for geode_object, value in geode_objects.geode_objects_dict().items():
            for input_extension in input_extensions:
                file_absolute_path = os.path.join(
                    data_folder, f"test.{input_extension}"
                )
                return_dict = geode_functions.list_geode_objects(
                    file_absolute_path, key
                )
                assert type(return_dict) is dict

                if key != None:
                    assert len(return_dict.keys()) > 0
                    for invalid_geode_object in invalid_geode_objects:
                        assert invalid_geode_object not in return_dict.keys()
                else:
                    assert len(return_dict.keys()) >= 1
                    for input_geode_object, input_extension_dict in return_dict.items():
                        assert type(input_extension_dict) is dict
                        for (
                            input_extension,
                            input_extension_value,
                        ) in input_extension_dict.items():
                            assert type(input_extension_value) is bool


def test_geode_objects_output_extensions():
    for geode_object, value in geode_objects.geode_objects_dict().items():
        input_extensions = geode_functions.geode_object_input_extensions(geode_object)
        for input_extension in input_extensions:
            file_absolute_path = os.path.join(data_folder, f"test.{input_extension}")
            if geode_functions.is_loadable(geode_object, file_absolute_path):
                data = geode_functions.load(geode_object, file_absolute_path)
                geode_objects_and_output_extensions = (
                    geode_functions.geode_objects_output_extensions(geode_object, data)
                )
                assert type(geode_objects_and_output_extensions) is dict
                for (
                    output_geode_object,
                    output_geode_object_value,
                ) in geode_objects_and_output_extensions.items():
                    for (
                        output_extension,
                        output_extension_value,
                    ) in output_geode_object_value.items():
                        assert type(output_extension_value["is_saveable"]) is bool



# Standard library imports

# Third party imports
import opengeode as og
import opengeode_io as og_io
import opengeode_inspector as og_inspector
import opengeode_geosciences as og_gs
import opengeode_geosciencesio as og_gs_io
import geode_viewables as g_v

# Local application imports


def geode_objects_dict():
    return {
        "BRep": {
            "class": og.BRep,
            "input_factory": og.BRepInputFactory,
            "output_factory": og.BRepOutputFactory,
            "missing_files": og.check_brep_missing_files,
            "is_loadable": og.is_brep_loadable,
            "load": og.load_brep,
            "is_saveable": og.is_brep_saveable,
            "save": og.save_brep,
            "builder": og.BRepBuilder,
            "crs": {
                "assign": og_gs.assign_brep_geographic_coordinate_system_info,
                "convert": og_gs.convert_brep_coordinate_reference_system,
                "create": og.create_brep_coordinate_system,
            },
            "object_type": "model",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_brep,
            "inspector": og_inspector.inspect_brep,
        },
        "CrossSection": {
            "parent": "Section",
            "class": og_gs.CrossSection,
            "input_factory": og_gs.CrossSectionInputFactory,
            "output_factory": og_gs.CrossSectionOutputFactory,
            "missing_files": og_gs.check_cross_section_missing_files,
            "is_loadable": og_gs.is_cross_section_loadable,
            "load": og_gs.load_cross_section,
            "is_saveable": og_gs.is_cross_section_saveable,
            "save": og_gs.save_cross_section,
            "builder": og_gs.CrossSectionBuilder,
            "crs": {
                "assign": og_gs.assign_section_geographic_coordinate_system_info,
                "convert": og_gs.convert_section_coordinate_reference_system,
                "create": og.create_section_coordinate_system,
            },
            "object_type": "model",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_cross_section,
            "inspector": og_inspector.inspect_section,
        },
        "EdgedCurve2D": {
            "class": og.EdgedCurve2D,
            "input_factory": og.EdgedCurveInputFactory2D,
            "output_factory": og.EdgedCurveOutputFactory2D,
            "missing_files": og.check_edged_curve_missing_files2D,
            "is_loadable": og.is_edged_curve_loadable2D,
            "load": og.load_edged_curve2D,
            "is_saveable": og.is_edged_curve_saveable2D,
            "save": og.save_edged_curve2D,
            "builder": og.EdgedCurveBuilder2D.create,
            "crs": {
                "assign": og_gs.assign_edged_curve_geographic_coordinate_system_info2D,
                "convert": og_gs.convert_edged_curve_coordinate_reference_system2D,
                "create": og.create_edged_curve_coordinate_system2D,
            },
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_edged_curve2D,
            "inspector": og_inspector.inspect_edged_curve2D,
        },
        "EdgedCurve3D": {
            "class": og.EdgedCurve3D,
            "input_factory": og.EdgedCurveInputFactory3D,
            "output_factory": og.EdgedCurveOutputFactory3D,
            "missing_files": og.check_edged_curve_missing_files3D,
            "is_loadable": og.is_edged_curve_loadable3D,
            "load": og.load_edged_curve3D,
            "is_saveable": og.is_edged_curve_saveable3D,
            "save": og.save_edged_curve3D,
            "builder": og.EdgedCurveBuilder3D.create,
            "crs": {
                "assign": og_gs.assign_edged_curve_geographic_coordinate_system_info3D,
                "convert": og_gs.convert_edged_curve_coordinate_reference_system3D,
                "create": og.create_edged_curve_coordinate_system3D,
            },
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_edged_curve3D,
            "inspector": og_inspector.inspect_edged_curve3D,
        },
        "Graph": {
            "class": og.Graph,
            "input_factory": og.GraphInputFactory,
            "output_factory": og.GraphOutputFactory,
            "missing_files": og.check_graph_missing_files,
            "is_loadable": og.is_graph_loadable,
            "load": og.load_graph,
            "is_saveable": og.is_graph_saveable,
            "save": og.save_graph,
            "builder": og.GraphBuilder.create,
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": True,
        },
        "HybridSolid3D": {
            "class": og.HybridSolid3D,
            "input_factory": og.HybridSolidInputFactory3D,
            "output_factory": og.HybridSolidOutputFactory3D,
            "missing_files": og.check_hybrid_solid_missing_files3D,
            "is_loadable": og.is_hybrid_solid_loadable3D,
            "load": og.load_hybrid_solid3D,
            "is_saveable": og.is_hybrid_solid_saveable3D,
            "save": og.save_hybrid_solid3D,
            "builder": og.HybridSolidBuilder3D.create,
            "crs": {
                "assign": og_gs.assign_solid_mesh_geographic_coordinate_system_info3D,
                "convert": og_gs.convert_solid_mesh_coordinate_reference_system3D,
                "create": og.create_solid_mesh_coordinate_system3D,
            },
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_hybrid_solid3D,
            "inspector": og_inspector.inspect_solid3D,
        },
        "ImplicitCrossSection": {
            "parent": "CrossSection",
            "class": og_gs.ImplicitCrossSection,
            "input_factory": og_gs.ImplicitCrossSectionInputFactory,
            "output_factory": og_gs.ImplicitCrossSectionOutputFactory,
            "missing_files": og_gs.check_implicit_cross_section_missing_files,
            "is_loadable": og_gs.is_implicit_cross_section_loadable,
            "load": og_gs.load_implicit_cross_section,
            "is_saveable": og_gs.is_implicit_cross_section_saveable,
            "save": og_gs.save_implicit_cross_section,
            "builder": og_gs.ImplicitCrossSectionBuilder,
            "crs": {
                "assign": og_gs.assign_section_geographic_coordinate_system_info,
                "convert": og_gs.convert_section_coordinate_reference_system,
                "create": og.create_section_coordinate_system,
            },
            "object_type": "model",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_implicit_cross_section,
            "inspector": og_inspector.inspect_section,
        },
        "ImplicitStructuralModel": {
            "parent": "StructuralModel",
            "class": og_gs.ImplicitStructuralModel,
            "input_factory": og_gs.ImplicitStructuralModelInputFactory,
            "output_factory": og_gs.ImplicitStructuralModelOutputFactory,
            "missing_files": og_gs.check_implicit_structural_model_missing_files,
            "is_loadable": og_gs.is_implicit_structural_model_loadable,
            "load": og_gs.load_implicit_structural_model,
            "is_saveable": og_gs.is_implicit_structural_model_saveable,
            "save": og_gs.save_implicit_structural_model,
            "builder": og_gs.ImplicitStructuralModelBuilder,
            "crs": {
                "assign": og_gs.assign_brep_geographic_coordinate_system_info,
                "convert": og_gs.convert_brep_coordinate_reference_system,
                "create": og.create_brep_coordinate_system,
            },
            "object_type": "model",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_implicit_structural_model,
            "inspector": og_inspector.inspect_brep,
        },
        "LightRegularGrid2D": {
            "class": og.LightRegularGrid2D,
            "input_factory": og.LightRegularGridInputFactory2D,
            "output_factory": og.LightRegularGridOutputFactory2D,
            "missing_files": og.check_light_regular_grid_missing_files2D,
            "is_loadable": og.is_light_regular_grid_loadable2D,
            "load": og.load_light_regular_grid2D,
            "is_saveable": og.is_light_regular_grid_saveable2D,
            "save": og.save_light_regular_grid2D,
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_light_regular_grid2D,
        },
        "LightRegularGrid3D": {
            "class": og.LightRegularGrid3D,
            "input_factory": og.LightRegularGridInputFactory3D,
            "output_factory": og.LightRegularGridOutputFactory3D,
            "missing_files": og.check_light_regular_grid_missing_files3D,
            "is_loadable": og.is_light_regular_grid_loadable3D,
            "load": og.load_light_regular_grid3D,
            "is_saveable": og.is_light_regular_grid_saveable3D,
            "save": og.save_light_regular_grid3D,
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_light_regular_grid3D,
        },
        "PointSet2D": {
            "class": og.PointSet2D,
            "input_factory": og.PointSetInputFactory2D,
            "output_factory": og.PointSetOutputFactory2D,
            "missing_files": og.check_point_set_missing_files2D,
            "is_loadable": og.is_point_set_loadable2D,
            "load": og.load_point_set2D,
            "is_saveable": og.is_point_set_saveable2D,
            "save": og.save_point_set2D,
            "builder": og.PointSetBuilder2D.create,
            "crs": {
                "assign": og_gs.assign_point_set_geographic_coordinate_system_info2D,
                "convert": og_gs.convert_point_set_coordinate_reference_system2D,
                "create": og.create_point_set_coordinate_system2D,
            },
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_point_set2D,
            "inspector": og_inspector.inspect_point_set2D,
        },
        "PointSet3D": {
            "class": og.PointSet3D,
            "input_factory": og.PointSetInputFactory3D,
            "output_factory": og.PointSetOutputFactory3D,
            "missing_files": og.check_point_set_missing_files3D,
            "is_loadable": og.is_point_set_loadable3D,
            "load": og.load_point_set3D,
            "is_saveable": og.is_point_set_saveable3D,
            "save": og.save_point_set3D,
            "builder": og.PointSetBuilder3D.create,
            "crs": {
                "assign": og_gs.assign_point_set_geographic_coordinate_system_info3D,
                "convert": og_gs.convert_point_set_coordinate_reference_system3D,
                "create": og.create_point_set_coordinate_system3D,
            },
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_point_set3D,
            "inspector": og_inspector.inspect_point_set3D,
        },
        "PolygonalSurface2D": {
            "class": og.PolygonalSurface2D,
            "input_factory": og.PolygonalSurfaceInputFactory2D,
            "output_factory": og.PolygonalSurfaceOutputFactory2D,
            "missing_files": og.check_polygonal_surface_missing_files2D,
            "is_loadable": og.is_polygonal_surface_loadable2D,
            "load": og.load_polygonal_surface2D,
            "is_saveable": og.is_polygonal_surface_saveable2D,
            "save": og.save_polygonal_surface2D,
            "builder": og.PolygonalSurfaceBuilder2D.create,
            "crs": {
                "assign": og_gs.assign_surface_mesh_geographic_coordinate_system_info2D,
                "convert": og_gs.convert_surface_mesh_coordinate_reference_system2D,
                "create": og.create_surface_mesh_coordinate_system2D,
            },
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_polygonal_surface2D,
            "inspector": og_inspector.inspect_surface2D,
        },
        "PolygonalSurface3D": {
            "class": og.PolygonalSurface3D,
            "input_factory": og.PolygonalSurfaceInputFactory3D,
            "output_factory": og.PolygonalSurfaceOutputFactory3D,
            "missing_files": og.check_polygonal_surface_missing_files3D,
            "is_loadable": og.is_polygonal_surface_loadable3D,
            "load": og.load_polygonal_surface3D,
            "is_saveable": og.is_polygonal_surface_saveable3D,
            "save": og.save_polygonal_surface3D,
            "builder": og.PolygonalSurfaceBuilder3D.create,
            "crs": {
                "assign": og_gs.assign_surface_mesh_geographic_coordinate_system_info3D,
                "convert": og_gs.convert_surface_mesh_coordinate_reference_system3D,
                "create": og.create_surface_mesh_coordinate_system3D,
            },
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_polygonal_surface3D,
            "inspector": og_inspector.inspect_surface3D,
        },
        "PolyhedralSolid3D": {
            "class": og.PolyhedralSolid3D,
            "input_factory": og.PolyhedralSolidInputFactory3D,
            "output_factory": og.PolyhedralSolidOutputFactory3D,
            "missing_files": og.check_polyhedral_solid_missing_files3D,
            "is_loadable": og.is_polyhedral_solid_loadable3D,
            "load": og.load_polyhedral_solid3D,
            "is_saveable": og.is_polyhedral_solid_saveable3D,
            "save": og.save_polyhedral_solid3D,
            "builder": og.PolyhedralSolidBuilder3D.create,
            "crs": {
                "assign": og_gs.assign_solid_mesh_geographic_coordinate_system_info3D,
                "convert": og_gs.convert_solid_mesh_coordinate_reference_system3D,
                "create": og.create_solid_mesh_coordinate_system3D,
            },
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_polyhedral_solid3D,
            "inspector": og_inspector.inspect_solid3D,
        },
        "RasterImage2D": {
            "class": og.RasterImage2D,
            "input_factory": og.RasterImageInputFactory2D,
            "output_factory": og.RasterImageOutputFactory2D,
            "missing_files": og.check_raster_image_missing_files2D,
            "is_loadable": og.is_raster_image_loadable2D,
            "load": og.load_raster_image2D,
            "is_saveable": og.is_raster_image_saveable2D,
            "save": og.save_raster_image2D,
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": False,
            "save_viewable": g_v.save_viewable_raster_image2D,
        },
        "RasterImage3D": {
            "class": og.RasterImage3D,
            "input_factory": og.RasterImageInputFactory3D,
            "output_factory": og.RasterImageOutputFactory3D,
            "missing_files": og.check_raster_image_missing_files3D,
            "is_loadable": og.is_raster_image_loadable3D,
            "load": og.load_raster_image3D,
            "is_saveable": og.is_raster_image_saveable3D,
            "save": og.save_raster_image3D,
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": False,
            "save_viewable": g_v.save_viewable_raster_image3D,
        },
        "RegularGrid2D": {
            "class": og.RegularGrid2D,
            "input_factory": og.RegularGridInputFactory2D,
            "output_factory": og.RegularGridOutputFactory2D,
            "missing_files": og.check_regular_grid_missing_files2D,
            "is_loadable": og.is_regular_grid_loadable2D,
            "load": og.load_regular_grid2D,
            "is_saveable": og.is_regular_grid_saveable2D,
            "save": og.save_regular_grid2D,
            "builder": og.RegularGridBuilder2D.create,
            "crs": {
                "assign": og_gs.assign_surface_mesh_geographic_coordinate_system_info2D,
                "convert": og_gs.convert_surface_mesh_coordinate_reference_system2D,
                "create": og.create_surface_mesh_coordinate_system2D,
            },
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_regular_grid2D,
        },
        "RegularGrid3D": {
            "class": og.RegularGrid3D,
            "input_factory": og.RegularGridInputFactory3D,
            "output_factory": og.RegularGridOutputFactory3D,
            "missing_files": og.check_regular_grid_missing_files3D,
            "is_loadable": og.is_regular_grid_loadable3D,
            "load": og.load_regular_grid3D,
            "is_saveable": og.is_regular_grid_saveable3D,
            "save": og.save_regular_grid3D,
            "builder": og.RegularGridBuilder3D.create,
            "crs": {
                "assign": og_gs.assign_solid_mesh_geographic_coordinate_system_info3D,
                "convert": og_gs.convert_solid_mesh_coordinate_reference_system3D,
                "create": og.create_surface_mesh_coordinate_system3D,
            },
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_regular_grid3D,
        },
        "Section": {
            "class": og.Section,
            "input_factory": og.SectionInputFactory,
            "output_factory": og.SectionOutputFactory,
            "missing_files": og.check_section_missing_files,
            "is_loadable": og.is_section_loadable,
            "load": og.load_section,
            "is_saveable": og.is_section_saveable,
            "save": og.save_section,
            "builder": og.SectionBuilder,
            "crs": {
                "assign": og_gs.assign_section_geographic_coordinate_system_info,
                "convert": og_gs.convert_section_coordinate_reference_system,
                "create": og.create_section_coordinate_system,
            },
            "object_type": "model",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_section,
            "inspector": og_inspector.inspect_section,
        },
        "StructuralModel": {
            "parent": "BRep",
            "class": og_gs.StructuralModel,
            "input_factory": og_gs.StructuralModelInputFactory,
            "output_factory": og_gs.StructuralModelOutputFactory,
            "missing_files": og_gs.check_structural_model_missing_files,
            "is_loadable": og_gs.is_structural_model_loadable,
            "load": og_gs.load_structural_model,
            "is_saveable": og_gs.is_structural_model_saveable,
            "save": og_gs.save_structural_model,
            "builder": og_gs.StructuralModelBuilder,
            "crs": {
                "assign": og_gs.assign_brep_geographic_coordinate_system_info,
                "convert": og_gs.convert_brep_coordinate_reference_system,
                "create": og.create_brep_coordinate_system,
            },
            "object_type": "model",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_structural_model,
            "inspector": og_inspector.inspect_brep,
        },
        "TetrahedralSolid3D": {
            "class": og.TetrahedralSolid3D,
            "input_factory": og.TetrahedralSolidInputFactory3D,
            "output_factory": og.TetrahedralSolidOutputFactory3D,
            "missing_files": og.check_tetrahedral_solid_missing_files3D,
            "is_loadable": og.is_tetrahedral_solid_loadable3D,
            "load": og.load_tetrahedral_solid3D,
            "is_saveable": og.is_tetrahedral_solid_saveable3D,
            "save": og.save_tetrahedral_solid3D,
            "builder": og.TetrahedralSolidBuilder3D.create,
            "crs": {
                "assign": og_gs.assign_solid_mesh_geographic_coordinate_system_info3D,
                "convert": og_gs.convert_solid_mesh_coordinate_reference_system3D,
                "create": og.create_solid_mesh_coordinate_system3D,
            },
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_tetrahedral_solid3D,
            "inspector": og_inspector.inspect_solid3D,
        },
        "TriangulatedSurface2D": {
            "class": og.TriangulatedSurface2D,
            "input_factory": og.TriangulatedSurfaceInputFactory2D,
            "output_factory": og.TriangulatedSurfaceOutputFactory2D,
            "missing_files": og.check_triangulated_surface_missing_files2D,
            "is_loadable": og.is_triangulated_surface_loadable2D,
            "load": og.load_triangulated_surface2D,
            "is_saveable": og.is_triangulated_surface_saveable2D,
            "save": og.save_triangulated_surface2D,
            "builder": og.TriangulatedSurfaceBuilder2D.create,
            "crs": {
                "assign": og_gs.assign_surface_mesh_geographic_coordinate_system_info2D,
                "convert": og_gs.convert_surface_mesh_coordinate_reference_system2D,
                "create": og.create_surface_mesh_coordinate_system2D,
            },
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_triangulated_surface2D,
            "inspector": og_inspector.inspect_surface2D,
        },
        "TriangulatedSurface3D": {
            "class": og.TriangulatedSurface3D,
            "input_factory": og.TriangulatedSurfaceInputFactory3D,
            "output_factory": og.TriangulatedSurfaceOutputFactory3D,
            "missing_files": og.check_triangulated_surface_missing_files3D,
            "is_loadable": og.is_triangulated_surface_loadable3D,
            "load": og.load_triangulated_surface3D,
            "is_saveable": og.is_triangulated_surface_saveable3D,
            "save": og.save_triangulated_surface3D,
            "builder": og.TriangulatedSurfaceBuilder3D.create,
            "crs": {
                "assign": og_gs.assign_surface_mesh_geographic_coordinate_system_info3D,
                "convert": og_gs.convert_surface_mesh_coordinate_reference_system3D,
                "create": og.create_surface_mesh_coordinate_system3D,
            },
            "object_type": "mesh",
            "is_3D": True,
            "is_viewable": True,
            "save_viewable": g_v.save_viewable_triangulated_surface3D,
            "inspector": og_inspector.inspect_surface3D,
        },
        "VertexSet": {
            "class": og.VertexSet,
            "input_factory": og.VertexSetInputFactory,
            "output_factory": og.VertexSetOutputFactory,
            "missing_files": og.check_vertex_set_missing_files,
            "is_loadable": og.is_vertex_set_loadable,
            "load": og.load_vertex_set,
            "is_saveable": og.is_vertex_set_saveable,
            "save": og.save_vertex_set,
            "builder": og.VertexSetBuilder.create,
            "object_type": "mesh",
            "is_3D": False,
            "is_viewable": False,
        },
    }
